﻿// Anyone is free to copy, modify, use, compile, or distribute this software,
// either in source code form or as a compiled binary, for any non-commercial
// purpose.  For commercial enquiries please contact contact@gimpysoftware.com
//
// If modified, please retain this copyright header, and consider telling us
// about your changes.  We're always glad to see how people use our code!
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
// Gimpy Software does not accept any liability for damage caused by executing
// or modifying this code.
//
// Copyright (c) 2013, Gimpy Software.
//
using System.Collections.Generic;
using System.Diagnostics;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace Speculator
{
    public class ZXDisplay
    {
        internal const int ScreenBase = 0x4000;
        private const int ColorMapBase = 0x5800;

        private static readonly List<Color> colors = new List<Color>
                                                     {
                                                         Color.FromRgb(0x00, 0x00, 0x00),  // Black
                                                         Color.FromRgb(0x00, 0x00, 0xCD),  // Blue
                                                         Color.FromRgb(0xCD, 0x00, 0x00),  // Red
                                                         Color.FromRgb(0xCD, 0x00, 0xCD),  // Magenta
                                                         Color.FromRgb(0x00, 0xCD, 0x00),  // Green
                                                         Color.FromRgb(0x00, 0xCD, 0xCD),  // Cyan
                                                         Color.FromRgb(0xCD, 0xCD, 0x00),  // Yellow
                                                         Color.FromRgb(0xCD, 0xCD, 0xCD),  // White (/Gray)
                                                         Color.FromRgb(0x00, 0x00, 0x00),  // (Bright) Black
                                                         Color.FromRgb(0x00, 0x00, 0xFF),  // Bright Blue
                                                         Color.FromRgb(0xFF, 0x00, 0x00),  // Bright Red
                                                         Color.FromRgb(0xFF, 0x00, 0xFF),  // Bright Magenta
                                                         Color.FromRgb(0x00, 0xFF, 0x00),  // Bright Green
                                                         Color.FromRgb(0x00, 0xFF, 0xFF),  // Bright Cyan
                                                         Color.FromRgb(0xFF, 0xFF, 0x00),  // Bright Yellow
                                                         Color.FromRgb(0xFF, 0xFF, 0xFF)   // Bright White
                                                     };

        private const int LeftMargin = 32;
        private const int RightMargin = 32;
        private const int TopMargin = 24;
        private const int BottomMargin = 24;
        private const int WriteableWidth = 256;
        private const int WritableHeight = 192;

        public WriteableBitmap Bitmap { get; private set; }

        public ZXDisplay()
        {
            Bitmap = new WriteableBitmap(LeftMargin + WriteableWidth + RightMargin, TopMargin + WritableHeight + BottomMargin, 96, 96, PixelFormats.Indexed8, new BitmapPalette(colors));
        }

        private unsafe void SetContentPixelGroup(int characterColumn, int y, byte pixels, byte attr, bool invertColors)
        {
            int x = characterColumn * 8;

            Debug.Assert(x >= 0 && x <= Bitmap.PixelWidth);
            Debug.Assert(y >= 0 && y <= Bitmap.PixelHeight);

            // Get a pointer to the back buffer.
            int pBackBuffer = (int)Bitmap.BackBuffer;

            // Find the address of the pixel to draw.
            pBackBuffer += y * Bitmap.BackBufferStride;
            pBackBuffer += x;

            // Assign the color data to the pixel.
            int penIndex;
            int paperIndex;
            if (invertColors)
                GetColorIndeces(attr, out penIndex, out paperIndex);
            else
                GetColorIndeces(attr, out paperIndex, out penIndex);

            for (int i = 7; i >= 0; i--)
            {
                bool isSet = (pixels & (1 << i)) != 0;
                *((byte*)pBackBuffer) = (byte)(isSet ? penIndex : paperIndex);
                pBackBuffer++;
            }
        }

        private static void GetColorIndeces(byte attr, out int paperIndex, out int penIndex)
        {
            paperIndex = (attr >> 3) & 0x07;
            penIndex = attr & 0x07;
        }

        private void Lock()
        {
            Bitmap.Lock();
        }

        private void Unlock()
        {
            Bitmap.AddDirtyRect(new Int32Rect(0, 0, Bitmap.PixelWidth, Bitmap.PixelHeight));
            Bitmap.Unlock();
        }

        private byte m_previousBorderColor;

        private const int FramesPerFlash = 16;
        private int m_flashFrameCount;
        private bool m_isFlashing;
        internal void UpdateScreen(CPU sender)
        {
            try
            {
                Lock();

                byte borderAttr = sender.MainMemory.Peek(0x5C48);

                if (m_previousBorderColor != borderAttr)
                {
                    m_previousBorderColor = borderAttr;
                    // Draw top/bottom borders.
                    for (int x = 0; x < Bitmap.PixelWidth / 8; x++)
                    {
                        for (int y = 0; y < TopMargin; y++)
                            SetContentPixelGroup(x, y, 0x00, borderAttr, false);
                        for (int y = 0; y < BottomMargin; y++)
                            SetContentPixelGroup(x, TopMargin + WritableHeight + y, 0x00, borderAttr, false);
                    }

                    // Draw left/right borders.
                    for (int y = 0; y < WritableHeight; y++)
                    {
                        for (int x = 0; x < LeftMargin / 8; x++)
                            SetContentPixelGroup(x, TopMargin + y, 0x00, borderAttr, false);
                        for (int x = 0; x < RightMargin / 8; x++)
                            SetContentPixelGroup(((LeftMargin + WriteableWidth) / 8) + x, TopMargin + y, 0x00, borderAttr, false);
                    }
                }

                // Flashing?
                if (m_flashFrameCount++ == FramesPerFlash)
                {
                    m_isFlashing = !m_isFlashing;
                    m_flashFrameCount = 0;
                }

                // Draw content.
                for (int i = 0; i < 6144; i++)
                {
                    int addr = ScreenBase + i;
                    int tt = (addr >> 11) & 0x03;
                    int lll = (addr >> 8) & 0x07;
                    int cr = (addr >> 5) & 0x07;
                    int cc = addr & 0x1F;

                    int pixelY = (cr * 8) + (tt * 64) + lll;
                    byte screenByte = sender.MainMemory.Data[addr];

                    byte characterAttr = sender.MainMemory.Data[ColorMapBase + (((tt * 8) + cr) * 32) + cc];
                    bool isFlashSet = (characterAttr & 0x80) != 0;
                    SetContentPixelGroup(cc + (LeftMargin / 8), pixelY + TopMargin, screenByte, characterAttr, m_isFlashing && isFlashSet);
                }
            }
            finally
            {
                Unlock();
            }
        }
    }
}
