﻿// Anyone is free to copy, modify, use, compile, or distribute this software,
// either in source code form or as a compiled binary, for any non-commercial
// purpose.  For commercial enquiries please contact contact@gimpysoftware.com
//
// If modified, please retain this copyright header, and consider telling us
// about your changes.  We're always glad to see how people use our code!
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
// Gimpy Software does not accept any liability for damage caused by executing
// or modifying this code.
//
// Copyright (c) 2013, Gimpy Software.
//
using System;
using System.Windows;
using System.Windows.Threading;
using Speculator;

// ReSharper disable UnusedMember.Global
// ReSharper disable MemberCanBePrivate.Global
namespace SpeculatorApp
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow
    {
        private readonly RegistersWindow m_registersWindow;
        private readonly ZxSpectrum m_theSpectrum;
        private readonly DispatcherTimer m_oneSecondTimer;
        private KeyboardWindow m_keyboardWindow;

        public MainWindow()
        {
            InitializeComponent();
            DataContext = this;

            DisplayControl.DisplaySource = new ZXDisplay();

            m_theSpectrum = new ZxSpectrum(DisplayControl.DisplaySource)
                            {
                                TheCpu = { IsDebugging = false }
                            };
            IsThrottledTo100Percent = true;
            m_theSpectrum.LoadBasicRom();

            m_registersWindow = new RegistersWindow(m_theSpectrum);

            m_theSpectrum.TheCpu.PowerOnAsync();

            Closed += MainWindow_Closed;

            KeyDown += OnKeyDown;
            KeyUp += OnKeyUp;
            Deactivated += ClearKeyPresses;
            Closed += delegate { m_theSpectrum.Dispose(); };

            m_oneSecondTimer = new DispatcherTimer { Interval = TimeSpan.FromSeconds(1.0) };
            m_oneSecondTimer.Tick += OnOneSecondElapsed;
            m_oneSecondTimer.Start();
        }

        void ClearKeyPresses(object sender, EventArgs eventArgs)
        {
            m_theSpectrum.PortHandler.ClearKeys();
        }

        void OnKeyDown(object sender, System.Windows.Input.KeyEventArgs e)
        {
            m_theSpectrum.PortHandler.SetKeyDown(e.Key);
        }

        void OnKeyUp(object sender, System.Windows.Input.KeyEventArgs e)
        {
            m_theSpectrum.PortHandler.SetKeyUp(e.Key);
        }

        private void OnOneSecondElapsed(object sender, EventArgs e)
        {
            Title = string.Format("Speculator: Running at {0} FPS", (int)Math.Round(m_theSpectrum.TheCpu.EmulationFPS));
        }

        private void MainWindow_Closed(object sender, EventArgs e)
        {
            m_theSpectrum.TheCpu.PowerOffAsync();
            if (m_registersWindow != null) m_registersWindow.Close();
            if (m_keyboardWindow != null) m_keyboardWindow.Close();
        }

        public bool IsThrottledTo100Percent
        {
            get
            {
                return !m_theSpectrum.TheCpu.FullThrottle;
            }

            set
            {
                m_theSpectrum.TheCpu.FullThrottle = !value;
            }
        }

        public bool IsDebuggerVisible
        {
            get { return m_registersWindow.Visibility == Visibility.Visible; }

            set { m_registersWindow.Visibility = value ? Visibility.Visible : Visibility.Hidden; }
        }

        public bool IsKeyboardVisible
        {
            get { return m_keyboardWindow != null; }

            set
            {
                if (value)
                {
                    m_keyboardWindow = new KeyboardWindow();
                    m_keyboardWindow.Width = m_keyboardWindow.KeyboardImage.Width;
                    m_keyboardWindow.Height = m_keyboardWindow.KeyboardImage.Height;
                    m_keyboardWindow.Show();
                    m_keyboardWindow.Closed += delegate { m_keyboardWindow = null; };
                }
                else
                {
                    m_keyboardWindow.Hide();
                    m_keyboardWindow = null;
                }
            }
        }
    }
}
